let myhero = require('../../myhero');
let Package = require('myhero-protocol').Package;
let crypto = require('crypto');

let CODE_OK = 200;
let CODE_USE_ERROR = 500;
let CODE_OLD_CLIENT = 501;

let DEFAULT_HEARTBEAT_IN_SEC = 3; //默认的心跳间隔时长
let DEFAULT_AES_KEY = 'XKWK-AES-CBC-128'; //默认的aes值
let DEFAULT_CBCIV = 'FFA514765D6B7A6C'; //默认的aes iv， aes_key的md5的16位摘要值
let DEFAULT_ALGORITHM = 'aes-128-cbc';
let CHAR_STR = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ==';

/**
 * Process the handshake request.
 *
 * @param {Object} opts option parameters
 *                      opts.handshake(msg, cb(err, resp)) handshake callback. msg is the handshake message from client.
 *                      opts.hearbeat heartbeat interval (level?)
 *                      opts.version required client level
 */
let Command = function (opts) {
    opts = opts || {};
    this.userHandshake = opts.handshake;

    // if (opts.heartbeat) {
    this.heartbeatSec = opts.heartbeat || DEFAULT_HEARTBEAT_IN_SEC;
    this.heartbeat = this.heartbeatSec * 1000;
    // }

    this.checkClient = opts.checkClient;

    this.useDict = opts.useDict;
    this.useProtobuf = opts.useProtobuf;
    this.useCrypto = opts.useCrypto;

    this.aesKey = opts.aesKey || DEFAULT_AES_KEY;   // AES 秘钥
    this.cbciv = opts.cbciv || DEFAULT_CBCIV;       // AES-128-CBC偏移量
    this.algorithm = opts.algorithm || DEFAULT_ALGORITHM; // 加密算法
};

module.exports = Command;

Command.prototype.handle = function (socket, msg) {
    if (!msg.sys) {
        processError(socket, CODE_USE_ERROR);
        return;
    }

    if (typeof this.checkClient === 'function') {
        if (!msg || !msg.sys || !this.checkClient(msg.sys.type, msg.sys.version, msg.sys.sha)) {
            processError(socket, CODE_OLD_CLIENT);
            return;
        }
    }

    let opts = {
        startupTime: getStartupTime(),
        heartbeat: setupHeartbeat(this)
    };

    if (this.useDict) {
        let dictVersion = myhero.app.components.__dictionary__.getVersion();
        if (!msg.sys.dictVersion || msg.sys.dictVersion !== dictVersion) {
            // may be deprecated in future
            opts.dicts = myhero.app.components.__dictionary__.getDicts();
        }
        opts.useDict = true;
    }

    if (this.useProtobuf) {
        let protoVersion = myhero.app.components.__protobuf__.getVersion();
        if (!msg.sys.protoVersion || msg.sys.protoVersion !== protoVersion) {
            opts.protos = myhero.app.components.__protobuf__.getProtos();
        }
        opts.useProto = true;
    }

    // if (!!myhero.app.components.__decodeIO__protobuf__) {
    //     if (!!this.useProtobuf) {
    //         throw new Error('protobuf can not be both used in the same project.');
    //     }
    //     let version = myhero.app.components.__decodeIO__protobuf__.getVersion();
    //     if (!msg.sys.protoVersion || msg.sys.protoVersion < version) {
    //         opts.protos = myhero.app.components.__decodeIO__protobuf__.getProtos();
    //     }
    //     opts.useProto = true;
    // }


    let pubKey = new Buffer(generateRandomString()).toString('base64');
    if (this.useCrypto) {
        myhero.app.components.__connector__.setPubKey(socket.id, pubKey);
        opts.pubKey = pubKey;
    }

    let encrypted = encryption(JSON.stringify(opts), this.aesKey, this.cbciv, this.algorithm);

    if (typeof this.userHandshake === 'function') {
        this.userHandshake(msg, function (err, resp) {
            if (err) {
                setImmediate(function () {
                    processError(socket, CODE_USE_ERROR);
                });
                return;
            }
            setImmediate(function () {
                response(socket, encrypted, resp);
            });
        }, socket);
        return;
    }

    setImmediate(function () {
        response(socket, encrypted);
    });
};

/**
 * 生产一个随机的32位字符长度的字符串
 */
let generateRandomString = function () {
    let randomString = [];
    for (let i = 0; i < 32; i++) {
        randomString.push(CHAR_STR.charAt(Math.floor(Math.random() * 64)));
    }
    return randomString.join('');
};

/**
 * aes-128-cbc加密
 */
let encryption = function (data, key, iv, algorithm) {
    let clearEncoding = 'utf8';
    let cipherEncoding = 'base64';
    let cipherChunks = [];
    let cipher = crypto.createCipheriv(algorithm, key, iv);
    cipher.setAutoPadding(true);
    cipherChunks.push(cipher.update(data, clearEncoding, cipherEncoding));
    cipherChunks.push(cipher.final(cipherEncoding));
    return cipherChunks.join('');
};

// all servers startup timestamp
let getStartupTime = function () {
    let app = require('../../myhero').app;
    var startupTime = app.get('startupTime');
    if (!startupTime) {
        startupTime = app.startTime;
    }
    return startupTime;
};

let setupHeartbeat = function (self) {
    return self.heartbeatSec;
};

let response = function (socket, sys, resp) {
    let res = {
        code: CODE_OK,
        sys: sys
    };
    if (resp) {
        res.user = resp;
    }
    socket.handshakeResponse(Package.encode(Package.TYPE_HANDSHAKE, new Buffer(JSON.stringify(res))));
};

let processError = function (socket, code) {
    let res = {
        code: code,
        msg: 'handshake failed: ' + (code === 500 ? 'user error' : 'old client version')
    };
    socket.sendForce(Package.encode(Package.TYPE_HANDSHAKE, new Buffer(JSON.stringify(res))));
    setImmediate(function () {
        socket.disconnect();
    });
};
